@@ -4,7 +4,7 @@ This module contains coba configuration functionality.
 """
 
 from coba.config.core    import CobaConfig
-from coba.config.cachers import NullCacher, MemoryCacher, DiskCacher, Cacher
+from coba.config.cachers import NullCacher, MemoryCacher, DiskCacher, Cacher, ConcurrentCacher
 from coba.config.loggers import NullLogger, BasicLogger, IndentLogger, Logger
 
 __all__ =[
@@ -12,6 +12,7 @@ __all__ =[
     'NullCacher',
     'MemoryCacher',
     'DiskCacher',
+    'ConcurrentCacher',
     'Cacher',
     'NullLogger',
     'BasicLogger',
