@@ -1,57 +1,40 @@
-import collections
+import collections.abc
 
 from math import isnan
 from statistics import mean, median, stdev, mode
 from numbers import Number
+from collections import defaultdict
 from abc import abstractmethod, ABC
 from itertools import islice, chain
 from typing_extensions import Literal
-from typing import Hashable, Optional, Sequence, cast, Union, Iterable, Dict, Any, List, Tuple
+from typing import Hashable, Optional, Sequence, Union, Iterable, Dict, Any, List, Tuple
 
 from coba.random import CobaRandom
 from coba.pipes import Filter
 from coba.statistics import iqr
 
-from coba.environments.core import SimulatedInteraction, LoggedInteraction, Interaction
+from coba.environments.primitives import SimulatedInteraction, LoggedInteraction, Interaction, EnvironmentFilter
 
-class SimulationFilter(Filter[Iterable[SimulatedInteraction],Iterable[SimulatedInteraction]], ABC):
+class Sort(EnvironmentFilter):
 
-    @property
-    @abstractmethod
-    def params(self) -> Dict[str, Any]:
-        ...
-
-    @abstractmethod
-    def filter(self, interactions: Iterable[SimulatedInteraction]) -> Iterable[SimulatedInteraction]:
-        """Apply a filter to a Simulation's interactions."""
-        ...
-
-    def __repr__(self) -> str:
-        return str(self.params)
-
-class Sort(SimulationFilter):
-
-    def __init__(self, *indexes: Union[int,Sequence[int]]) -> None:
+    def __init__(self, *keys: Union[str,int,Sequence[Union[str,int]]]) -> None:
         
-        flat_indexes = cast(Sequence[int], indexes[0] if isinstance(indexes[0], collections.Sequence) else indexes)
-
-        if not isinstance(flat_indexes, collections.Sequence) or not isinstance(flat_indexes[0],int):
-            raise ValueError(f"Invalid parameter for Sort: {flat_indexes}. A sequence of integers was expected.")
-
-        self._indexes = flat_indexes
+        self._keys = []
+        
+        for key in keys:
+            if not isinstance(key, collections.abc.Sequence) or isinstance(key,str):
+                self._keys.append(key)
+            else:
+                self._keys.extend(key)
 
     @property
     def params(self) -> Dict[str, Any]:
-        return { "sort": self._indexes }
+        return { "sort": self._keys }
 
-    def filter(self, interactions: Iterable[SimulatedInteraction]) -> Iterable[SimulatedInteraction]:
-        
-        return sorted(interactions, key=lambda interaction: tuple(interaction.context[i] for i in self._indexes))
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
+        return sorted(interactions, key=lambda interaction: tuple(interaction.context[key] for key in self._keys))
 
-    def __repr__(self) -> str:
-        return str(self.params)
-
-class Scale(SimulationFilter):
+class Scale(EnvironmentFilter):
 
     def __init__(self, 
         shift: Union[Number,Literal["min","mean","med"]] ="min", 
@@ -69,15 +52,15 @@ class Scale(SimulationFilter):
     def params(self) -> Dict[str, Any]:
         return { "scale_shift": self._shift, "scale_scale":self._scale, "scale_using":self._using }
 
-    def filter(self, interactions: Iterable[SimulatedInteraction]) -> Iterable[SimulatedInteraction]:
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
         iter_interactions  = iter(interactions)
         train_interactions = list(islice(iter_interactions,self._using))
         test_interactions  = chain.from_iterable([train_interactions, iter_interactions])
         
-        shifts  : Dict[Hashable,float]     = collections.defaultdict(lambda:0)
-        scales  : Dict[Hashable,float]     = collections.defaultdict(lambda:1)
-        features: Dict[Hashable,List[Any]] = collections.defaultdict(list)
+        shifts  : Dict[Hashable,float]     = defaultdict(lambda:0)
+        scales  : Dict[Hashable,float]     = defaultdict(lambda:1)
+        features: Dict[Hashable,List[Any]] = defaultdict(list)
 
         for interaction in train_interactions:
             for name,value in self._context_as_name_values(interaction.context):
@@ -86,9 +69,6 @@ class Scale(SimulationFilter):
 
         for feat_name, feat_numeric_values in features.items():
 
-            if isinstance(feat_numeric_values[0],str):
-                continue 
-
             if isinstance(self._shift, Number):
                 shifts[feat_name] = self._shift
 
@@ -134,11 +114,14 @@ class Scale(SimulationFilter):
             elif isinstance(interaction.context,dict):
                 final_context = kv_scaled_context
             elif isinstance(interaction.context,tuple):
-                final_context = tuple( kv_scaled_context[k] for k,_ in self._context_as_name_values(interaction.context))
+                final_context = tuple(kv_scaled_context[k] for k,_ in self._context_as_name_values(interaction.context))
             else:
                 final_context = kv_scaled_context[1]
 
-            yield SimulatedInteraction(final_context, interaction.actions, **interaction.kwargs)
+            try:
+                yield SimulatedInteraction(final_context, interaction.actions, **interaction.kwargs)
+            except:
+                yield LoggedInteraction(final_context, interaction.action, **interaction.kwargs)
 
     def _context_as_name_values(self,context) -> Sequence[Tuple[Hashable,Any]]:
         
@@ -148,35 +131,7 @@ class Scale(SimulationFilter):
 
         return []
 
-    def __repr__(self) -> str:
-        return str(self.params)
-
-class Cycle(SimulationFilter):
-
-    def __init__(self, after:int = 0):
-        self._after = after
-
-    @property
-    def params(self) -> Dict[str, Any]:
-        return { "cycle_after": self._after }
-
-    def filter(self, interactions: Iterable[SimulatedInteraction]) -> Iterable[SimulatedInteraction]:
-
-        underlying_iterable     = iter(interactions)
-        sans_cycle_interactions = islice(underlying_iterable, self._after)
-        with_cycle_interactions = underlying_iterable
-
-        for interaction in sans_cycle_interactions:
-            yield interaction
-
-        for interaction in with_cycle_interactions:
-            kwargs = {k:v[1:]+v[:1] for k,v in interaction.kwargs.items()}
-            yield SimulatedInteraction(interaction.context, interaction.actions, **kwargs)
-
-    def __repr__(self) -> str:
-        return str(self.params)
-
-class Impute(SimulationFilter):
+class Impute(EnvironmentFilter):
 
     def __init__(self, 
         stat : Literal["mean","median","mode"] = "mean",
@@ -191,14 +146,14 @@ class Impute(SimulationFilter):
     def params(self) -> Dict[str, Any]:
         return { "impute_stat": self._stat, "impute_using": self._using }
 
-    def filter(self, interactions: Iterable[SimulatedInteraction]) -> Iterable[SimulatedInteraction]:
+    def filter(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
 
         iter_interactions  = iter(interactions)
         train_interactions = list(islice(iter_interactions,self._using))
         test_interactions  = chain.from_iterable([train_interactions, iter_interactions])
         
-        stats   : Dict[Hashable,float]     = collections.defaultdict(lambda:0)
-        features: Dict[Hashable,List[Any]] = collections.defaultdict(list)
+        stats   : Dict[Hashable,float]        = defaultdict(int)
+        features: Dict[Hashable,List[Number]] = defaultdict(list)
 
         for interaction in train_interactions:
             for name,value in self._context_as_name_values(interaction.context):
@@ -207,9 +162,6 @@ class Impute(SimulationFilter):
 
         for feat_name, feat_numeric_values in features.items():
 
-            if isinstance(feat_numeric_values[0],str):
-                continue 
-
             if self._stat == "mean":
                 stats[feat_name] = mean(feat_numeric_values)
 
@@ -231,11 +183,14 @@ class Impute(SimulationFilter):
             elif isinstance(interaction.context,dict):
                 final_context = kv_imputed_context
             elif isinstance(interaction.context,tuple):
-                final_context = tuple( kv_imputed_context[k] for k,_ in self._context_as_name_values(interaction.context))
+                final_context = tuple(kv_imputed_context[k] for k,_ in self._context_as_name_values(interaction.context))
             else:
                 final_context = kv_imputed_context[1]
 
-            yield SimulatedInteraction(final_context, interaction.actions, **interaction.kwargs)
+            try:
+                yield SimulatedInteraction(final_context, interaction.actions, **interaction.kwargs)
+            except:
+                yield LoggedInteraction(final_context, interaction.action, **interaction.kwargs)
 
     def _context_as_name_values(self,context) -> Sequence[Tuple[Hashable,Any]]:
         
@@ -245,10 +200,29 @@ class Impute(SimulationFilter):
 
         return []
 
-    def __repr__(self) -> str:
-        return str(self.params)
+class Cycle(EnvironmentFilter):
 
-class Binary(SimulationFilter):
+    def __init__(self, after:int = 0):
+        self._after = after
+
+    @property
+    def params(self) -> Dict[str, Any]:
+        return { "cycle_after": self._after }
+
+    def filter(self, interactions: Iterable[SimulatedInteraction]) -> Iterable[SimulatedInteraction]:
+
+        underlying_iterable     = iter(interactions)
+        sans_cycle_interactions = islice(underlying_iterable, self._after)
+        with_cycle_interactions = underlying_iterable
+
+        for interaction in sans_cycle_interactions:
+            yield interaction
+
+        for interaction in with_cycle_interactions:
+            kwargs = {k:v[1:]+v[:1] for k,v in interaction.kwargs.items()}
+            yield SimulatedInteraction(interaction.context, interaction.actions, **kwargs)
+
+class Binary(EnvironmentFilter):
     @property
     def params(self) -> Dict[str, Any]:
         return { "binary": True }
@@ -262,9 +236,9 @@ class Binary(SimulationFilter):
 
             yield SimulatedInteraction(interaction.context, interaction.actions, **kwargs)
 
-class ToWarmStart(SimulationFilter):
+class ToWarmStart(EnvironmentFilter):
+
     def __init__(self, n_warmstart:int, seed:int = 1):
-        
         self._n_warmstart = n_warmstart
         self._seed = seed
 
@@ -282,16 +256,20 @@ class ToWarmStart(SimulationFilter):
 
         return chain(logged_interactions, simulated_interactions)
 
-    def _to_logged_interaction(self, interaction: SimulatedInteraction) -> LoggedInteraction:        
+    def _to_logged_interaction(self, interaction: SimulatedInteraction) -> LoggedInteraction:
         num_actions   = len(interaction.actions)
         probabilities = [1/num_actions] * num_actions 
         
         selected_index       = self._rng.choice(list(range(num_actions)), probabilities)
         selected_action      = interaction.actions[selected_index]
         selected_probability = probabilities[selected_index]
-        selected_reward      = interaction.kwargs.get("reveals", interaction.kwargs.get("rewards", None))[selected_index]
+        
+        kwargs = {"probability":selected_probability, "actions":interaction.actions}
 
-        return LoggedInteraction(interaction.context, selected_action, selected_reward, selected_probability)
+        if "reveals" in interaction.kwargs:
+            kwargs["reveal"] = interaction.kwargs["reveals"][selected_index]
 
-    def __repr__(self) -> str:
-        return str(self.params)
+        if "rewards" in interaction.kwargs:
+            kwargs["reward"] = interaction.kwargs["rewards"][selected_index]
+
+        return LoggedInteraction(interaction.context, selected_action, **kwargs)
