@@ -1,8 +1,10 @@
+import time
 import json
 
 from hashlib import md5
 from numbers import Number
 from collections import defaultdict
+from typing_extensions import Literal
 from typing import Tuple, Sequence, Any, Iterable, Dict, Union
 
 from coba.pipes import Pipe, Source, HttpIO, Default, Drop, Encode, _T_Data, Structure, ArffReader, CsvReader, Take
@@ -10,7 +12,8 @@ from coba.config import CobaConfig
 from coba.exceptions import CobaException
 from coba.encodings import NumericEncoder, OneHotEncoder, StringEncoder
 
-from coba.environments.simulations import SimulatedEnvironment, SimulatedInteraction, ClassificationSimulation, RegressionSimulation
+from coba.environments.primitives import SimulatedEnvironment, SimulatedInteraction
+from coba.environments.simulations import ClassificationSimulation, RegressionSimulation
 
 class OpenmlSource(Source[Union[Iterable[Tuple[_T_Data, str]], Iterable[Tuple[_T_Data, Number]]]]):
 
@@ -20,19 +23,25 @@ class OpenmlSource(Source[Union[Iterable[Tuple[_T_Data, str]], Iterable[Tuple[_T
 
         self._data_id      = id
         self._md5_checksum = md5_checksum
-        self._problem_type = problem_type 
+        self._problem_type = problem_type
         self._cat_as_str   = cat_as_str
         self._take         = take
-        self._cached_urls  = []
+        self._cache_keys   = {
+            'descr': f"openml_{id:0>6}_descr",
+            'feats': f"openml_{id:0>6}_feats",
+            'csv'  : f"openml_{id:0>6}_csv",
+            'arff' : f"openml_{id:0>6}_arff",
+            'tasks': f"openml_{id:0>6}_tasks",
+        }
 
     @property
     def params(self) -> Dict[str, Any]:
         """Paramaters describing the environment."""
 
         if self._take is not None:
-            return { "openml": self._data_id, "cat_as_str": self._cat_as_str, "openml_take": self._take }
+            return { "openml": self._data_id, "cat_as_str": self._cat_as_str, "openml_type": self._problem_type, "openml_take": self._take,  }
         else:
-            return { "openml": self._data_id, "cat_as_str": self._cat_as_str, }
+            return { "openml": self._data_id, "cat_as_str": self._cat_as_str, "openml_type": self._problem_type, }
 
     def read(self) -> Union[Iterable[Tuple[_T_Data, str]], Iterable[Tuple[_T_Data, Number]]]:
 
@@ -88,7 +97,7 @@ class OpenmlSource(Source[Union[Iterable[Tuple[_T_Data, str]], Iterable[Tuple[_T
             if self._problem_type == "classification":
                 encoders[target] = StringEncoder() if self._cat_as_str else OneHotEncoder()
 
-            file_rows = self._get_dataset_rows(dataset_description["file_id"], target, md5_checksum)
+            file_rows = self._get_dataset_rows(dataset_description["file_id"], md5_checksum)
 
             def row_has_missing_values(row):
                 row_values = row.values() if isinstance(row,dict) else row
@@ -105,127 +114,125 @@ class OpenmlSource(Source[Union[Iterable[Tuple[_T_Data, str]], Iterable[Tuple[_T
         except KeyboardInterrupt:
             #we don't want to clear the cache in the case of a KeyboardInterrupt
             raise
-        
+
         except CobaException:
+            #we don't want to clear the cache if it is an error we know about (the original raise should clear if needed)
             raise
 
         except Exception:
-            #if something unexpected went wrong clear the
-            #cache just in case it was corrupted somehow
-            
-            for key in self._cached_urls:
+            #if something unexpected went wrong clear the cache just in case it was corrupted somehow
+
+            for key in self._cache_keys.values():
                 CobaConfig.cacher.rmv(key)
 
             raise
 
-    def _get_url(self, url:str, checksum:str=None) -> Iterable[str]:
+    def _get_data(self, url:str, key:str, checksum:str=None) -> Iterable[str]:
         
-        if url in CobaConfig.cacher:
-            self._cached_urls.append(url)
-            bites = CobaConfig.cacher.get(url)
-        else:
+        # This can't reasonably be done in a streaming manner unless cacher is persistent.
+        # For now we don't require cacher to be consistent so I'm commenting out for now. 
+        # if checksum is not None and md5(bites).hexdigest() != checksum:
+        #     #if the cache has become corrupted we need to clear it
+        #     CobaConfig.cacher.rmv(key)
+        #     message = (
+        #         f"The response from {url} did not match the given checksum {checksum}. This could be the result "
+        #         "of network errors or the file becoming corrupted. Please consider downloading the file again. "
+        #         "If the error persists you may want to manually download and reference the file.")
+        #     raise CobaException(message) from None
 
-            api_key  = CobaConfig.api_keys['openml']
-            response = HttpIO(url + (f'?api_key={api_key}' if api_key else '')).read()
+        for b in CobaConfig.cacher.get_put(key, lambda: self._http_request(url)):
+            yield b.decode('utf-8')
 
-            if response.status_code == 412:
-                if 'please provide api key' in response.text:
+    def _http_request(self, url:str) -> Iterable[bytes]:
+        api_key  = CobaConfig.api_keys['openml']
+        
+        srcsema = CobaConfig.store.get("srcsema")
+        if srcsema: srcsema.acquire() # we only allow three paralellel request, another attempt at being more "considerate".
+            
+        # An attempt to be considerate of how often we hit their REST api. 
+        # They don't publish any rate-limiting guidelines so this is just a guess.
+        time.sleep(1) 
+
+        try:
+            with HttpIO(url + (f'?api_key={api_key}' if api_key else '')).read() as response:
+
+                if response.status_code == 412: # pragma: no cover
+                    if 'please provide api key' in response.text:
+                        message = (
+                            "Openml has requested an API Key to access openml's rest API. A key can be obtained by creating "
+                            "an openml account at openml.org. Once a key has been obtained it should be placed within "
+                            "~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")
+                        raise CobaException(message)
+
+                    if 'authentication failed' in response.text:
+                        message = (
+                            "The API Key you provided no longer seems to be valid. You may need to create a new one by "
+                            "logging into your openml account and regenerating a key. After regenerating the new key "
+                            "should be placed in ~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")
+                        raise CobaException(message)
+
+                if response.status_code == 404: # pragma: no cover
                     message = (
-                        "An API Key is needed to access openml's rest API. A key can be obtained by creating an "
-                        "openml account at openml.org. Once a key has been obtained it should be placed within "
-                        "~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")
-                    raise CobaException(message) from None
+                        "We're sorry but we were unable to find the requested dataset on openml.")
+                    raise CobaException(message)
 
-                if 'authentication failed' in response.text:
-                    message = (
-                        "The API Key you provided no longer seems to be valid. You may need to create a new one by "
-                        "logging into your openml account and regenerating a key. After regenerating the new key "
-                        "should be placed in ~/.coba as { \"api_keys\" : { \"openml\" : \"<your key here>\", } }.")
-                    raise CobaException(message) from None
+                # NOTE: These two checks need to be gated with a status code failure 
+                # NOTE: otherwise this will cause the data to be downloaded all at once
+                # NOTE: unfortunately I don't know the appropriate status code so commenting out for now
+                # if "Usually due to high server load" in response.text:
+                #     message = (
+                #         "Openml has experienced an error that they believe is the result of high server loads."
+                #         "Openml recommends that you try again in a few seconds. Additionally, if not already "
+                #         "done, consider setting up a DiskCache in CobaConfig to reduce the number of openml "
+                #         "calls in the future.")
+                #     raise CobaException(message) from None
 
-            if response.status_code == 404:
-                message = (
-                    "We're sorry but we were unable to find the requested dataset on openml.")
-                raise CobaException(message) from None
+                # if '' == response.text: # pragma: no cover
+                #     raise CobaException("Openml experienced an unexpected error. Please try requesting the data again.") from None
 
-            if "Usually due to high server load" in response.text:
-                message = (
-                    "Openml has experienced an error that they believe is the result of high server loads."
-                    "Openml recommends that you try again in a few seconds. Additionally, if not already "
-                    "done, consider setting up a DiskCache in CobaConfig to reduce the number of openml "
-                    "calls in the future.")
-                raise CobaException(message) from None
-
-            if '' == response.text:
-                raise CobaException("Openml experienced an unexpected error. Please try requesting the data again.") from None
-
-            bites = response.iter_lines(decode_unicode=False)
-
-            if url not in CobaConfig.cacher:
-                CobaConfig.cacher.put(url,bites)
-                
-                if url in CobaConfig.cacher:
-                    self._cached_urls.append(url)
-                    bites = CobaConfig.cacher.get(url)
-
-        if checksum is not None and md5(bites).hexdigest() != checksum:
-
-            #if the cache has become corrupted we need to clear it
-            CobaConfig.cacher.rmv(url)
-
-            message = (
-                f"The response from {url} did not match the given checksum {checksum}. This could be the result "
-                "of network errors or the file becoming corrupted. Please consider downloading the file again. "
-                "If the error persists you may want to manually download and reference the file.")
-            raise CobaException(message) from None
-
-        return ( b.decode('utf-8') for b in bites )
+                for b in response.iter_lines(decode_unicode=False):
+                    yield b
+        finally:
+            if srcsema: srcsema.release()
 
     def _get_dataset_description(self, data_id:int) -> Dict[str,Any]:
-
-        description_txt = " ".join(self._get_url(f'https://www.openml.org/api/v1/json/data/{data_id}'))
+        
+        description_txt = " ".join(self._get_data(f'https://www.openml.org/api/v1/json/data/{data_id}', self._cache_keys['descr']))
         description_obj = json.loads(description_txt)["data_set_description"]
 
         return description_obj
 
     def _get_feature_descriptions(self, data_id:int) -> Sequence[Dict[str,Any]]:
 
-        types_txt = " ".join(self._get_url(f'https://www.openml.org/api/v1/json/data/features/{data_id}'))
+        types_txt = " ".join(self._get_data(f'https://www.openml.org/api/v1/json/data/features/{data_id}', self._cache_keys['feats']))
         types_obj = json.loads(types_txt)["data_features"]["feature"]
 
         return types_obj
 
-    def _get_dataset_rows(self, file_id:str, target:str, md5_checksum:str) -> Any:
+    def _get_dataset_rows(self, file_id:str, md5_checksum:str) -> Any:
 
-            csv_url  = f"http://www.openml.org/data/v1/get_csv/{file_id}"
-            arff_url = f"http://www.openml.org/data/v1/download/{file_id}"
+            csv_url  = f"https://www.openml.org/data/v1/get_csv/{file_id}"
+            arff_url = f"https://www.openml.org/data/v1/download/{file_id}"
+
+            csv_key  = self._cache_keys['csv']
+            arff_key = self._cache_keys['arff']
 
             openml_dialect = dict(quotechar="'", escapechar="\\", doublequote=False)
 
-            if arff_url in CobaConfig.cacher:
-                return ArffReader(skip_encoding=True, **openml_dialect).filter(self._get_url(arff_url, md5_checksum))
-            
-            if csv_url in CobaConfig.cacher:
-                return CsvReader(True, **openml_dialect).filter(self._get_url(csv_url, md5_checksum))
-            
-            #we lock on downloading a dataset in case a user is running multithreaded and chunked by task
-            #in this case they could download the same data set repeatedly. This also makes requests a little
-            #nicer in terms of the load we place on openml's file server. srcsema should only be set when a
-            #user is running an experiment in a multithreaded environment.
-            
-            srcsema = CobaConfig.store.get("srcsema")
-            if srcsema: srcsema.acquire()
+            if arff_key in CobaConfig.cacher:
+                return ArffReader(skip_encoding=True, **openml_dialect).filter(self._get_data(arff_url, arff_key, md5_checksum))
+
+            if csv_key in CobaConfig.cacher:
+                return CsvReader(True, **openml_dialect).filter(self._get_data(csv_url, csv_key, md5_checksum))
 
             try:
-                return CsvReader(True, **openml_dialect).filter(self._get_url(csv_url, md5_checksum))
+                return CsvReader(True, **openml_dialect).filter(self._get_data(csv_url, csv_key, md5_checksum))
             except:
-                return ArffReader(skip_encoding=True, **openml_dialect).filter(self._get_url(arff_url, md5_checksum))
-            finally:
-                if srcsema: srcsema.release()
+                return ArffReader(skip_encoding=True, **openml_dialect).filter(self._get_data(arff_url, arff_key, md5_checksum))                
 
     def _get_target_for_problem_type(self, data_id:int):
 
-        text  = " ".join(self._get_url(f'https://www.openml.org/api/v1/json/task/list/data_id/{data_id}'))
+        text  = " ".join(self._get_data(f'https://www.openml.org/api/v1/json/task/list/data_id/{data_id}', self._cache_keys['tasks']))
         tasks = json.loads(text).get("tasks",{}).get("task",[])
 
         task_type = 1 if self._problem_type == "classification" else 2
@@ -239,10 +246,7 @@ class OpenmlSource(Source[Union[Iterable[Tuple[_T_Data, str]], Iterable[Tuple[_T
         raise CobaException(f"Openml {data_id} does not appear to be a {self._problem_type} dataset")
 
     def __repr__(self) -> str:
-        return f'{{"OpenmlSimulation":{self._data_id}}}'
-
-    def __str__(self) -> str:
-        return self.__repr__()
+        return f'{{"OpenmlSource":{self._data_id}}}'
 
 class OpenmlSimulation(SimulatedEnvironment):
     """A simulation created from openml data with features and labels.
@@ -255,8 +259,7 @@ class OpenmlSimulation(SimulatedEnvironment):
     incorrect lables).
     """
 
-    def __init__(self, id: int, take:int = None, simulation_type:str = "classification", cat_as_str:bool = False, md5_checksum: str = None) -> None:
-
+    def __init__(self, id: int, take:int = None, simulation_type:Literal["classification","regression"] = "classification", cat_as_str:bool = False, md5_checksum: str = None) -> None:
         self._sim_type = simulation_type
         self._source = OpenmlSource(id, simulation_type, cat_as_str, take, md5_checksum)
 
