@@ -1,5 +1,5 @@
 import math
-import collections
+import collections.abc
 
 from itertools import chain, repeat
 from typing import Sequence, Dict, Any, Iterable, Union, List, Callable, cast, Optional, Tuple, overload
@@ -7,11 +7,12 @@ from typing import Sequence, Dict, Any, Iterable, Union, List, Callable, cast, O
 from coba.pipes import Source, Filter, Pipe  
 from coba.pipes import HttpIO, DiskIO
 from coba.pipes import ResponseToLines, CsvReader, ArffReader, LibSvmReader, ManikReader, Structure
+from coba.pipes.io import MemoryIO
 from coba.random import CobaRandom
-from coba.encodings import InteractionTermsEncoder, OneHotEncoder, IdentityEncoder
+from coba.encodings import InteractionsEncoder, OneHotEncoder, IdentityEncoder
 from coba.exceptions import CobaException
 
-from coba.environments.core import Context, Action, SimulatedEnvironment, SimulatedInteraction
+from coba.environments.primitives import Context, Action, SimulatedEnvironment, SimulatedInteraction
 
 T_Dense_Feat  = Sequence[Any]
 T_Sparse_Feat = Dict[Any,Any]
@@ -27,7 +28,6 @@ class MemorySimulation(SimulatedEnvironment):
         Args:
             interactions: The sequence of interactions in this simulation.
         """
-
         self._interactions = interactions
 
     @property
@@ -70,13 +70,11 @@ class ClassificationSimulation(SimulatedEnvironment):
     def __init__(self, *args, **kwargs) -> None:
         """Instantiate a ClassificationSimulation."""
 
-        if len(args) == 1:
-            self._examples = args[0]
-        elif len(args) ==2:
-            self._examples = zip(args[0],args[1])
-        else:
+        if len(args) not in [1,2]:
             raise CobaException("We were unable to determine which overloaded constructor to use")
 
+        self._examples = args[0] if len(args) == 1 else zip(args[0],args[1])
+
     @property
     def params(self) -> Dict[str, Any]:
         """Paramaters describing the simulation."""
@@ -93,14 +91,14 @@ class ClassificationSimulation(SimulatedEnvironment):
 
             #how can we tell the difference between featurized labels and multilabels????
             #for now we will assume multilables will be passed in as arrays not tuples...
-            if not isinstance(labels[0], collections.Hashable):
+            if not isinstance(labels[0], collections.abc.Hashable):
                 labels_flat = list(chain.from_iterable(labels))
             else:
                 labels_flat = list(labels)
 
             reward        = lambda action,label: int(is_label(action,label) or in_multilabel(action,label)) #type: ignore
             is_label      = lambda action,label: action == label #type: ignore
-            in_multilabel = lambda action,label: isinstance(label,collections.Sequence) and action in label #type: ignore
+            in_multilabel = lambda action,label: isinstance(label,collections.abc.Sequence) and action in label #type: ignore
 
             # shuffling so that action order contains no statistical information
             # sorting so that the shuffled values are always shuffled in the same order
@@ -147,12 +145,10 @@ class RegressionSimulation(SimulatedEnvironment):
     def __init__(self, *args, **kwargs) -> None:
         """Instantiate a RegressionSimulation."""
 
-        if len(args) == 1:
-            self._examples = args[0]
-        elif len(args) == 2:
-            self._examples = zip(args[0],args[1])
-        else:
+        if len(args) not in [1,2]:
             raise CobaException("We were unable to determine which overloaded constructor to use")
+        
+        self._examples = args[0] if len(args) == 1 else zip(args[0],args[1])
 
     def read(self) -> Iterable[SimulatedInteraction]:
         """Read the interactions in this simulation."""
@@ -161,9 +157,9 @@ class RegressionSimulation(SimulatedEnvironment):
 
         if examples:
 
-            features,labels = zip(*self._examples)
+            features,labels = zip(*examples)
 
-            reward   = lambda action,label: 1-abs(float(action)-float(label))
+            reward   = lambda action,label: 1-round(abs(float(action)-float(label)),2)
             contexts = features
             actions  = CobaRandom(1).shuffle(sorted(set(labels)))
             rewards  = [ [ reward(action,label) for action in actions ] for label in labels ]
@@ -221,36 +217,27 @@ class ReaderSimulation(SimulatedEnvironment):
         source   : Union[str,Source[Iterable[str]]], 
         label_col: Union[str,int]) -> None:
         
-        self._reader = reader
-
-        if isinstance(source, str) and source.startswith('http'):
-            self._source = Pipe.join(HttpIO(source), [ResponseToLines()])
-        elif isinstance(source, str):
-            self._source = DiskIO(source)
-        else:
-            self._source = source
-
+        self._reader       = reader
+        self._source       = DiskIO(source) if isinstance(source, str) else source
         self._label_column = label_col
-        self._interactions = cast(Optional[Sequence[SimulatedInteraction]], None)
 
     @property
     def params(self) -> Dict[str, Any]:
         """Paramaters describing the simulation."""
-        return {"source": str(self._source) }
+        if isinstance(self._source,DiskIO):
+            return {"source": str(self._source.filename) }
+        elif isinstance(self._source,MemoryIO):
+            return {"source": 'memory' }
+        else:
+            return {"source": self._source.__class__.__name__}
 
     def read(self) -> Iterable[SimulatedInteraction]:
         """Read the interactions in this simulation."""
-        return self._load_interactions()
-
-    def _load_interactions(self) -> Sequence[SimulatedInteraction]:
         parsed_rows_iter = iter(self._reader.filter(self._source.read()))
         structured_rows = Structure([None, self._label_column]).filter(parsed_rows_iter)
 
         return ClassificationSimulation(structured_rows).read()
 
-    def __repr__(self) -> str:
-        return str(self.params)
-
 class CsvSimulation(ReaderSimulation):
     def __init__(self, source:Union[str,Source[Iterable[str]]], label_column:Union[str,int], with_header:bool=True) -> None:
         super().__init__(CsvReader(with_header), source, label_column)
@@ -280,7 +267,7 @@ class LibsvmSimulation(ReaderSimulation):
 
 class ManikSimulation(ReaderSimulation):
     def __init__(self, source:Union[str,Source[Iterable[str]]]) -> None:
-        super().__init__(ManikReader(), source, 0, False)
+        super().__init__(ManikReader(), source, 0)
 
     @property
     def params(self) -> Dict[str, Any]:
@@ -303,14 +290,14 @@ class DebugSimulation(LambdaSimulation):
         self._n_action_features  = n_context_feats
         self._seed               = seed
         self._r_noise_var        = r_noise_var
-        self._interactions       = interactions
+        self._X                  = interactions
 
         rng = CobaRandom(seed)
 
         action_encoder = OneHotEncoder([str(a) for a in range(n_actions)]) if n_action_feats == 0 else IdentityEncoder()
-        interactions_encoder = InteractionTermsEncoder(self._interactions)
+        interactions_encoder = InteractionsEncoder(self._X)
 
-        feature_count = len(interactions_encoder.encode(x=range(max(1,n_context_feats)),a=range(max(n_actions,n_action_feats))))
+        feature_count = len(interactions_encoder.encode(x=list(range(max(1,n_context_feats))),a=list(range(max(n_actions,n_action_feats)))))
         normalize     = lambda   X: [ rng.random()*x/sum(X) for x in X]
 
         weights = normalize(rng.randoms(feature_count))
@@ -322,7 +309,7 @@ class DebugSimulation(LambdaSimulation):
             return None if n_context_feats == 0 else rng.randoms(n_context_feats)
 
         def reward(index:int, context:Context, action:Action) -> float:
-            
+
             W = weights
             X = context or [1]
             A = action_encoder.encode([action])[0]
@@ -337,10 +324,15 @@ class DebugSimulation(LambdaSimulation):
     @property
     def params(self) -> Dict[str, Any]:
         """Paramaters describing the simulation."""
-        return { }
+
+        return { 
+            "|A|"     : self._n_actions,
+            "|phi(C)|": self._n_context_features,
+            "|phi(A)|": self._n_action_features,
+            "e_var"   : self._r_noise_var,
+            "X"       : self._X,
+            "seed"    : self._seed
+        }
 
     def __repr__(self) -> str:
-        return f"DebugSimulation(A={self._n_actions},c={self._n_context_features},a={self._n_action_features},seed={self._seed})"
-
-    def __str__(self) -> str:
-        return self.__repr__()
\ No newline at end of file
+        return f"DebugSimulation(A={self._n_actions},c={self._n_context_features},a={self._n_action_features},X={self._X},seed={self._seed})"
