@@ -2,17 +2,16 @@ from pathlib import Path
 from typing_extensions import Literal
 from typing import Sequence, Optional
 
+from coba.config import CobaConfig
+from coba.pipes import Pipe
 from coba.learners import Learner
 from coba.environments import Environment
-from coba.config import CobaConfig
-from coba.exceptions import CobaFatal
-from coba.pipes import Pipe
 from coba.multiprocessing import CobaMultiprocessFilter
 
-from coba.experiments.process      import CreateWorkItems,  RemoveFinished, ChunkByTask, ChunkBySource, ProcessWorkItems
-from coba.experiments.tasks        import EnvironmentTask, EvaluationTask, LearnerTask
-from coba.experiments.tasks        import SimpleLearnerTask, SimpleEnvironmentTask, OnPolicyEvaluationTask
-from coba.experiments.results      import Result, TransactionIO
+from coba.experiments.process import CreateWorkItems,  RemoveFinished, ChunkByTask, ChunkBySource, ProcessWorkItems
+from coba.experiments.tasks   import EnvironmentTask, EvaluationTask, LearnerTask
+from coba.experiments.tasks   import SimpleLearnerTask, SimpleEnvironmentTask, OnPolicyEvaluationTask
+from coba.experiments.results import Result, TransactionIO
 
 class Experiment:
     """An Experiment which uses environments and learners to calculate performance statistics."""
@@ -33,7 +32,7 @@ class Experiment:
             evaluation_task: A task which evaluates performance of a learner on an environment's interactions.
         """
 
-        self._environments      = environments
+        self._environments     = environments
         self._learners         = learners
         self._learner_task     = learner_task
         self._environment_task = environment_task
@@ -65,6 +64,18 @@ class Experiment:
 
         return self
 
+    @property
+    def chunk_by(self) -> str:
+        return self._chunk_by if self._chunk_by is not None else CobaConfig.experiment.chunk_by
+
+    @property
+    def processes(self) -> int:
+        return self._processes if self._processes is not None else CobaConfig.experiment.processes
+
+    @property
+    def maxtasksperchild(self) -> int:
+        return self._maxtasksperchild if self._maxtasksperchild is not None else CobaConfig.experiment.maxtasksperchild
+
     def evaluate(self, result_file:str = None) -> Result:
         """Evaluate the experiment and return the results.
 
@@ -81,9 +92,7 @@ class Experiment:
             assert n_given_learners     == restored.experiment['n_learners'    ], "The current experiment doesn't match the given transaction log."
             assert n_given_environments == restored.experiment['n_environments'], "The current experiment doesn't match the given transaction log."
 
-        cb = self._chunk_by         if self._chunk_by         is not None else CobaConfig.experiment.chunk_by
-        mp = self._processes        if self._processes        is not None else CobaConfig.experiment.processes
-        mt = self._maxtasksperchild if self._maxtasksperchild is not None else CobaConfig.experiment.maxtasksperchild
+        cb, mp, mt = self.chunk_by, self.processes, self.maxtasksperchild
 
         workitems  = CreateWorkItems(self._environments, self._learners, self._learner_task, self._environment_task, self._evaluation_task)
         unfinished = RemoveFinished(restored)
@@ -91,18 +100,17 @@ class Experiment:
         sink       = TransactionIO(result_file)
 
         single_process = ProcessWorkItems()
-        multi_process  = Pipe.join([chunk, CobaMultiprocessFilter([ProcessWorkItems()], mp, mt)])
+        multi_process  = Pipe.join([chunk, CobaMultiprocessFilter(ProcessWorkItems(), mp, mt)])
         process        = multi_process if mp > 1 or mt != 0 else single_process
 
         try:
             if not restored: sink.write([["T0", n_given_learners, n_given_environments]])
             Pipe.join(workitems, [unfinished, process], sink).run()
 
-        except KeyboardInterrupt:
+        except KeyboardInterrupt: # pragma: no cover
             CobaConfig.logger.log("Experiment execution was manually aborted via Ctrl-C")
-        except CobaFatal:
-            raise
-        except Exception as ex:
-            CobaConfig.logger.log_exception(ex)
+        
+        except Exception as ex: # pragma: no cover
+            CobaConfig.logger.log(ex)
 
         return sink.result
\ No newline at end of file
