@@ -1,6 +1,7 @@
 import gc
 
-from itertools import groupby, product
+from copy import deepcopy
+from itertools import groupby
 from collections import defaultdict
 from typing import Iterable, Sequence, Any, Optional, Tuple, Union
 
@@ -57,7 +58,7 @@ class CreateWorkItems(Source[Iterable[WorkItem]]):
 
         for env_id, env in keyed_environs.items():
             yield WorkItem(env_id, None, env, None, self._environment_task)
-            
+
         for env_id,env in keyed_environs.items():
             for lrn_id,lrn in keyed_learners.items():
                 yield WorkItem(env_id, lrn_id, env, lrn, self._evaluation_task)
@@ -90,7 +91,6 @@ class ChunkBySource(Filter[Iterable[WorkItem], Iterable[Sequence[WorkItem]]]):
 
         sans_source_items = [t for t in items if t.environ_id is None]
         with_source_items = [t for t in items if t.environ_id is not None]
-        
 
         for env_item in with_source_items:
             chunks[self._get_source(env_item.environ)].append(env_item)
@@ -100,7 +100,7 @@ class ChunkBySource(Filter[Iterable[WorkItem], Iterable[Sequence[WorkItem]]]):
 
         for chunk in sorted(chunks.values(), key=lambda chunk: min([c.environ_id for c in chunk])):
             yield list(sorted(chunk, key=lambda c: (c.environ_id, -1 if c.learner_id is None else c.learner_id)))
-    
+
     def _get_source(self, env):
         return env._source  if isinstance(env, (EnvironmentPipe, Pipe.SourceFilters)) else env
 
@@ -129,6 +129,13 @@ class ProcessWorkItems(Filter[Iterable[WorkItem], Iterable[Any]]):
 
         chunk = list(chunk)
 
+        self._source_id = {}
+
+        for item in chunk:
+            if item.environ_id is not None:
+                source = id(self._get_source(item))
+                self._source_id[source] = min(self._source_id.get(source,float('inf')), item.environ_id)
+
         if not chunk: return
 
         with CobaConfig.logger.log(f"Processing chunk..."):
@@ -140,8 +147,7 @@ class ProcessWorkItems(Filter[Iterable[WorkItem], Iterable[Any]]):
                     if env_source is None:
                         loaded_source = None
                     else:
-                        with CobaConfig.logger.time(f"Loading source {env_source}..."):
-                            
+                        with CobaConfig.logger.time(f"Loading {env_source}..."):
                             #This is not ideal. I'm not sure how it should be improved so it is being left for now.
                             #Maybe add a flag to the Experiment to say whether the source should be stashed in mem?
                             loaded_source = list(env_source.read())
@@ -153,7 +159,7 @@ class ProcessWorkItems(Filter[Iterable[WorkItem], Iterable[Any]]):
                         if loaded_source is None:
                             interactions = []
                         else:
-                            with CobaConfig.logger.time(f"Creating Environment {env_id} from {env_source}..."):
+                            with CobaConfig.logger.time(f"Creating Environment {env_id} from Loaded Source..."):
                                 interactions = list(env_filter.filter(loaded_source)) if env_filter else loaded_source
 
                             if len(filter_groups) == 1:
@@ -170,24 +176,27 @@ class ProcessWorkItems(Filter[Iterable[WorkItem], Iterable[Any]]):
 
                                 if workitem.environ is None:
                                     with CobaConfig.logger.time(f"Recording Learner {workitem.learner_id} parameters..."):
-                                        yield ["T1", workitem.learner_id, workitem.task.filter(workitem.learner)]
+                                        row = workitem.task.process(deepcopy(workitem.learner))
+                                        yield ["T1", workitem.learner_id, row]
 
                                 if workitem.learner is None:
                                     with CobaConfig.logger.time(f"Recording Environment {workitem.environ_id} statistics..."):
-                                        yield ["T2", workitem.environ_id, workitem.task.filter((workitem.environ,interactions))]
+                                        row = workitem.task.process(workitem.environ,interactions)
+                                        yield ["T2", workitem.environ_id, row]
 
                                 if workitem.environ and workitem.learner:
                                     with CobaConfig.logger.time(f"Evaluating Learner {workitem.learner_id} on Environment {workitem.environ_id}..."):
-                                        yield ["T3", (workitem.environ_id, workitem.learner_id), list(workitem.task.filter((workitem.learner, interactions)))]
+                                        row = list(workitem.task.process(deepcopy(workitem.learner), interactions))
+                                        yield ["T3", (workitem.environ_id, workitem.learner_id), row]
 
                             except Exception as e:
-                                CobaConfig.logger.log_exception(e)
+                                CobaConfig.logger.log(e)
 
                 except Exception as e:
-                    CobaConfig.logger.log_exception(e)
+                    CobaConfig.logger.log(e)
 
     def _get_source(self, task:WorkItem) -> SimulatedEnvironment:
-        if task.environ is None: 
+        if task.environ is None:
             return None
         elif isinstance(task.environ, (Pipe.SourceFilters, EnvironmentPipe)):
             return task.environ._source 
@@ -195,7 +204,7 @@ class ProcessWorkItems(Filter[Iterable[WorkItem], Iterable[Any]]):
             return task.environ
     
     def _get_source_sort(self, task:WorkItem) -> int:
-        return id(self._get_source(task))
+        return self._source_id.get(id(self._get_source(task)),-1)
 
     def _get_id_filter(self, task:WorkItem) -> Tuple[int, Filter[SimulatedEnvironment,SimulatedEnvironment]]:
         if task.environ is None:
