@@ -1,32 +1,32 @@
 from abc import ABC, abstractmethod
-from copy import deepcopy
 from statistics import median
 from collections import defaultdict
 from itertools import takewhile, chain
-from typing import Iterable, Any, Dict, Tuple
+from typing import Iterable, Any, Dict
 
+from coba.exceptions import CobaExit
 from coba.random import CobaRandom
 from coba.learners import Learner, SafeLearner
-from coba.pipes import Filter
 from coba.environments import Environment, EnvironmentPipe, Interaction, SimulatedInteraction, LoggedInteraction
-from coba.encodings import InteractionTermsEncoder
+from coba.encodings import InteractionsEncoder
+from coba.utilities import PackageChecker
 
-class LearnerTask(Filter[Learner,Dict[Any,Any]], ABC):
+class LearnerTask(ABC):
 
     @abstractmethod
-    def filter(self, item: Learner) -> Dict[Any,Any]:
+    def process(self, item: Learner) -> Dict[Any,Any]:
         ...
 
-class EnvironmentTask(Filter[Tuple[Environment,Iterable[Interaction]], Dict[Any,Any]]):
+class EnvironmentTask(ABC):
 
     @abstractmethod
-    def filter(self, item: Tuple[Environment,Iterable[Interaction]]) -> Dict[Any,Any]:
+    def process(self, environment: Environment, interactions: Iterable[Interaction]) -> Dict[Any,Any]:
         ...
 
-class EvaluationTask(Filter[Tuple[Learner,Iterable[Interaction]], Iterable[Dict[Any,Any]]]):
+class EvaluationTask(ABC):
 
     @abstractmethod
-    def filter(self, item: Tuple[Learner,Iterable[Interaction]]) -> Iterable[Dict[Any,Any]]:
+    def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Dict[Any,Any]]:
         ...
 
 class OnPolicyEvaluationTask(EvaluationTask):
@@ -34,11 +34,9 @@ class OnPolicyEvaluationTask(EvaluationTask):
     def __init__(self, seed:int = 1):
         self._seed = seed
 
-    def filter(self, item: Tuple[Learner,Iterable[SimulatedInteraction]]) -> Iterable[Dict[Any,Any]]:
+    def process(self, learner: Learner, interactions: Iterable[SimulatedInteraction]) -> Iterable[Dict[Any,Any]]:
 
-        learner      = deepcopy(item[0])
-        interactions = item[1]
-        random       = CobaRandom(self._seed)
+        random = CobaRandom(self._seed)
 
         if not isinstance(learner, SafeLearner): learner = SafeLearner(learner)
         if not interactions: return
@@ -63,24 +61,23 @@ class OnPolicyEvaluationTask(EvaluationTask):
 
 class OffPolicyEvaluationTask(EvaluationTask):
 
-    def filter(self, item: Tuple[Learner,Iterable[LoggedInteraction]]) -> Iterable[Dict[Any,Any]]:
-
-        learner      = deepcopy(item[0])
-        interactions = item[1]
+    def process(self, learner: Learner, interactions: Iterable[LoggedInteraction]) -> Iterable[Dict[Any,Any]]:
 
         if not isinstance(learner, SafeLearner): learner = SafeLearner(learner)
         if not interactions: return
 
         for interaction in interactions:
 
-            if not interaction.probability or not interaction.actions:
+            if len(interaction.kwargs.keys() & {"probability", "actions", "reward"}) != 3:
                 predict_info = {}
             else:
-                probs = learner.predict(interaction.context, interaction.actions)[0]
-                ratio = probs[interaction.actions.index(interaction.action)] / interaction.probability
-                predict_info = { "reward": ratio * interaction.reward }
+                actions      = list(interaction.kwargs["actions"])
+                probs        = learner.predict(interaction.context, actions)[0]
+                ratio        = probs[actions.index(interaction.action)] / interaction.kwargs["probability"]
+                predict_info = { "reward": ratio * interaction.kwargs["reward"] }
 
-            learn_info = learner.learn(interaction.context, interaction.action, interaction.reward, interaction.probability)
+            reveal     = interaction.kwargs.get("reveal", interaction.kwargs["reward"])
+            learn_info = learner.learn(interaction.context, interaction.action, reveal, interaction.kwargs.get("probability"))
 
             yield {**predict_info, **learn_info}
 
@@ -89,10 +86,7 @@ class WarmStartEvaluationTask(EvaluationTask):
     def __init__(self, seed: int) -> None:
         self._seed = seed
 
-    def filter(self, item: Tuple[Learner,Iterable[Interaction]]) -> Iterable[Dict[Any,Any]]:
-
-        learner      = deepcopy(item[0])
-        interactions = item[1]
+    def process(self, learner: Learner, interactions: Iterable[Interaction]) -> Iterable[Dict[Any,Any]]:
 
         if not isinstance(learner, SafeLearner): learner = SafeLearner(learner)
         if not interactions: return
@@ -102,10 +96,10 @@ class WarmStartEvaluationTask(EvaluationTask):
         logged_interactions    = takewhile(lambda i: isinstance(i,LoggedInteraction), separable_interactions)
         simulated_interactions = separable_interactions
 
-        for row in OffPolicyEvaluationTask().filter( (learner, logged_interactions)):
+        for row in OffPolicyEvaluationTask().process( (learner, logged_interactions)):
             yield row
 
-        for row in OnPolicyEvaluationTask(self._seed).filter( (learner, simulated_interactions)):
+        for row in OnPolicyEvaluationTask(self._seed).process( (learner, simulated_interactions)):
             yield row
 
     def _repeat_first_simulated(self, interactions: Iterable[Interaction]) -> Iterable[Interaction]:
@@ -116,20 +110,19 @@ class WarmStartEvaluationTask(EvaluationTask):
             else:
                 first_simulated_interaction = interaction
                 break
-        
+
         for interaction in chain([first_simulated_interaction], interactions):
             yield interaction
 
 class SimpleLearnerTask(LearnerTask):
-    def filter(self, item: Learner) -> Dict[Any,Any]:
+
+    def process(self, item: Learner) -> Dict[Any,Any]:
         item = SafeLearner(item)
         return {"full_name": item.full_name, **item.params}
 
 class SimpleEnvironmentTask(EnvironmentTask):
 
-    def filter(self, item: Tuple[Environment,Iterable[Interaction]]) -> Dict[Any,Any]:
-
-            environment  = item[0]
+    def process(self, environment:Environment, interactions: Iterable[Interaction]) -> Dict[Any,Any]:
 
             source = self._source_repr(environment)
             params = self._pipe_params(environment)
@@ -138,10 +131,10 @@ class SimpleEnvironmentTask(EnvironmentTask):
             
     def _source_repr(self, env) -> str:
         if isinstance(env, EnvironmentPipe):
-            return env.source_repr
+            return str(env._source)
         else:
             return env.__class__.__name__
-    
+
     def _pipe_params(self, env) -> Dict[str,Any]:
         if isinstance(env, EnvironmentPipe):
             return env.params
@@ -150,11 +143,15 @@ class SimpleEnvironmentTask(EnvironmentTask):
 
 class ClassEnvironmentTask(EnvironmentTask):
 
-    def filter(self, item: Tuple[Environment,Iterable[SimulatedInteraction]]) -> Dict[Any,Any]:
+    def process(self, environment: Environment, interactions: Iterable[SimulatedInteraction]) -> Dict[Any,Any]:
 
-        contexts,actions,rewards = zip(*[ (i.context, i.actions, i.kwargs["rewards"]) for i in item[1]])
+        contexts,actions,rewards = zip(*[ (i.context, i.actions, i.kwargs["rewards"]) for i in interactions ])
+        env_statistics = {}
 
         try:
+
+            PackageChecker.sklearn("ClassEnvironmentTask.process")
+
             import numpy as np
             import scipy.sparse as sp
             import scipy.stats as st
@@ -164,9 +161,7 @@ class ClassEnvironmentTask(EnvironmentTask):
             from sklearn.metrics import pairwise_distances
             from sklearn.decomposition import TruncatedSVD, PCA
 
-            env_statistics = {}
-
-            X   = [ InteractionTermsEncoder('x').encode(x=c, a=[]) for c in contexts ]
+            X   = [ InteractionsEncoder('x').encode(x=c, a=[]) for c in contexts ]
             Y   = [ a[r.index(1)] for a,r in zip(actions,rewards)]
             C   = defaultdict(list)
             clf = DecisionTreeClassifier(random_state=1)
@@ -197,10 +192,10 @@ class ClassEnvironmentTask(EnvironmentTask):
             closest_index  = centroid_dists.argmin(1)
             cluster_purity = (closest_index == centroid_index).mean()
 
-            env_statistics["centroid_purity"]   = round(cluster_purity,4)
+            env_statistics["centroid_purity"  ] = round(cluster_purity,4)
             env_statistics["centroid_distance"] = round(median(centroid_dists[range(centroid_dists.shape[0]),centroid_index]),4)
 
-        except ImportError:
+        except CobaExit:
             pass
 
         labels     = set()
@@ -223,4 +218,4 @@ class ClassEnvironmentTask(EnvironmentTask):
         env_statistics["context_median_nz" ] = median(feat_cnts)
         env_statistics["imbalance_ratio"   ] = round(max(label_cnts.values())/min(label_cnts.values()),4)
 
-        return { **SimpleEnvironmentTask().filter(item), **env_statistics }
+        return { **SimpleEnvironmentTask().process(environment, interactions), **env_statistics }
