@@ -1,11 +1,11 @@
-import collections
+import collections.abc
 import time
 
 from typing import Any, Dict, Sequence
 
 from coba.utilities import PackageChecker
 from coba.environments import Context, Action
-from coba.encodings import InteractionTermsEncoder
+from coba.encodings import InteractionsEncoder
 from coba.learners.core import Info, Learner
 
 class LinUCBLearner(Learner):
@@ -37,7 +37,7 @@ class LinUCBLearner(Learner):
         self._timeit       = timeit
 
         self._interactions = interactions
-        self._interactions_encoder = InteractionTermsEncoder(interactions)
+        self._interactions_encoder = InteractionsEncoder(interactions)
 
     def predict(self, context: Context, actions: Sequence[Action]) -> Sequence[float]:
         """Determine a PMF with which to select the given actions.
@@ -53,7 +53,7 @@ class LinUCBLearner(Learner):
 
         self._i += 1
 
-        self._d   = len(actions[0]) if isinstance(actions[0], collections.Sequence) else 1
+        self._d   = len(actions[0]) if isinstance(actions[0], collections.abc.Sequence) else 1
         is_sparse = isinstance(actions[0], dict) or isinstance(context, dict)
 
         if is_sparse:
@@ -119,4 +119,4 @@ class LinUCBLearner(Learner):
         for action in actions:
             features.append(self._interactions_encoder.encode(x=context,a=action))
         
-        return np.array(features).T
\ No newline at end of file
+        return np.array(features).T
