@@ -1,23 +1,22 @@
 """The vowpal module contains classes to make it easier to interact with pyvw."""
 
 import re
-import collections
+import collections.abc
 
-from os import devnull
 from itertools import repeat
 from numbers import Number
 from typing_extensions import Literal
 from typing import Any, Dict, Union, Sequence, overload, cast, Optional, Tuple
 
 from coba.exceptions import CobaException
-from coba.utilities import PackageChecker, redirect_stderr, KeyDefaultDict
+from coba.utilities import PackageChecker, KeyDefaultDict
 from coba.environments import Context, Action
 
 from coba.learners.core import Learner, Probs, Info
 
 Vowpal_Key      = Union[str,int]
-Vowpal_Features = Sequence[Union[str,Tuple[Vowpal_Key,Number]]]
-Coba_Feature    = Union[str,Number]
+Vowpal_Features = Sequence[Union[str,Tuple[Vowpal_Key,float]]]
+Coba_Feature    = Union[str,float]
 Coba_Features   = Union[Coba_Feature, Sequence[Coba_Feature], Sequence[Tuple[str,Coba_Feature]], Dict[str,Coba_Feature]]
 
 class VowpalMediator:
@@ -41,14 +40,14 @@ class VowpalMediator:
             return [("0",features)]
         elif isinstance(features,dict):
             features = features.items()
-        elif isinstance(features, collections.Sequence) and features and isinstance(features[0], tuple):
+        elif isinstance(features, collections.abc.Sequence) and features and isinstance(features[0], tuple):
             features = features
-        elif isinstance(features, collections.Sequence) and features and not isinstance(features[0],tuple):
+        elif isinstance(features, collections.abc.Sequence) and features and not isinstance(features[0],tuple):
             features = zip(map(VowpalMediator._string_cache.__getitem__, range(len(features))) ,features)
         else:
             raise Exception(f"Unrecognized features of type {type(features).__name__} passed to VowpalLearner.")
         
-        return [f"{F[0]}={F[1]}" if isinstance(F[1],str) else F for F in features if F[1] != 0]        
+        return [f"{F[0]}={F[1]}" if isinstance(F[1],str) else (F[0], float(F[1])) for F in features if F[1] != 0]        
 
     @staticmethod
     def make_learner(args:str):
@@ -268,14 +267,12 @@ class VowpalLearner(Learner):
         """
 
         if self._vw is None:
-             #type: ignore
             # vowpal has an annoying warning that is written to stderr whether or not we provide
             # the --quiet flag. Therefore, we temporarily redirect all stderr output to null so that
             # this warning isn't shown during creation. It should be noted this isn't thread-safe
             # so if you are here because of strange problems with threads we may just need to suck
             # it up and accept that there will be an obnoxious warning message.
-            with open(devnull, 'w') as f, redirect_stderr(f):
-                self._vw = VowpalMediator.make_learner(self._cli_args(actions) + " --quiet")
+            self._vw = VowpalMediator.make_learner(self._cli_args(actions) + " --quiet")
 
             if not self._adf:
                 self._actions = actions
