@@ -1,31 +1,11 @@
 """The data.pipes module contains core classes for creating data pipelines."""
 
-import collections
+import collections.abc
 
-from abc    import ABC, abstractmethod
-from typing import Sequence, Iterable, Any, overload, Union, TypeVar, Generic
+from typing import Sequence, Iterable, Any, overload, Union
 
-_T_out = TypeVar("_T_out", bound=Any, covariant=True    )
-_T_in  = TypeVar("_T_in" , bound=Any, contravariant=True)
-
-class StopPipe(Exception):
-    pass
-
-class Source(ABC, Generic[_T_out]):
-    @abstractmethod
-    def read(self) -> _T_out:
-        ...
-
-class Filter(ABC, Generic[_T_in, _T_out]):
-    @abstractmethod
-    def filter(self, item: _T_in) -> _T_out:
-        ...
-
-class Sink(ABC, Generic[_T_in]):
-
-    @abstractmethod
-    def write(self, item: _T_in) -> None:
-        ...
+from coba.pipes.primitives import Filter, Source, Sink, StopPipe
+from coba.pipes.filters import Identity 
 
 class Pipe:
 
@@ -121,9 +101,9 @@ class Pipe:
             return Pipe(*args)
 
         if len(args) == 2:
-            if isinstance(args[1], collections.Sequence):
+            if isinstance(args[1], collections.abc.Sequence):
                 return Pipe.SourceFilters(args[0], args[1])
-            elif isinstance(args[0], collections.Sequence):
+            elif isinstance(args[0], collections.abc.Sequence):
                 return Pipe.FiltersSink(args[0], args[1])
             else:
                 return Pipe(args[0], [], args[1])
