@@ -2,13 +2,13 @@
 
 import re
 import csv
-import collections
 import itertools
 import json
 import math
+import collections.abc
 
 from itertools import islice
-from typing_extensions import OrderedDict
+from collections import OrderedDict
 from typing import Iterable, Any, Sequence, Union, Tuple, List, Dict, Callable, Optional
 
 from requests import Response
@@ -16,7 +16,8 @@ from requests import Response
 from coba.random import CobaRandom
 from coba.encodings import Encoder, OneHotEncoder, NumericEncoder, StringEncoder, CobaJsonEncoder, CobaJsonDecoder
 from coba.exceptions import CobaException
-from coba.pipes.core import Filter
+
+from coba.pipes.primitives import Filter
 
 _T_DenseRow   = Sequence[Any]
 _T_SparseRow  = Dict[Any,Any]
@@ -30,11 +31,11 @@ class Cartesian(Filter[Union[Any,Iterable[Any]], Iterable[Any]]):
 
     def __init__(self, filter: Union[Filter,Sequence[Filter]]):
 
-        self._filters = filter if isinstance(filter, collections.Sequence) else [filter]
+        self._filters = filter if isinstance(filter, collections.abc.Sequence) else [filter]
 
     def filter(self, item: Union[Any,Iterable[Any]]) -> Iterable[Any]:
 
-        items = item if isinstance(item, collections.Iterable) else [item]
+        items = item if isinstance(item, collections.abc.Iterable) else [item]
 
         for item in items:
             for filter in self._filters:
@@ -69,12 +70,12 @@ class Shuffle(Filter[Iterable[Any], Iterable[Any]]):
 class Take(Filter[Iterable[Any], Iterable[Any]]):
     """Take a given number of items from an iterable."""
 
-    def __init__(self, count:Optional[int], keep_first:bool = False, seed: int = None,) -> None:
+    def __init__(self, count:Optional[int], seed: int = None, keep_first:bool = False) -> None:
         """Instantiate a Take filter.
 
         Args:
             count     : The number of items we wish to take from the given iterable.
-            keep_first: Indicates if the first row should be kept and take on the rest. Useful for files with headers.
+            keep_first: Indicates if the first row should be kept with take applied to the rest. Useful for files with headers.
             seed      : An optional random seed to determine which random count items to take.
 
         Remarks:
@@ -88,8 +89,8 @@ class Take(Filter[Iterable[Any], Iterable[Any]]):
         if count is not None and (not isinstance(count,int) or count < 0):
             raise ValueError(f"Invalid parameter for Take: {count}. An optional integer value >= 0 was expected.")
 
-        self._count = count
-        self._seed  = seed
+        self._count      = count
+        self._seed       = seed
         self._keep_first = keep_first
 
     @property
@@ -178,7 +179,10 @@ class JsonEncode(Filter[Any, str]):
                 elif math.isnan(v) or math.isinf(v):
                     obj[k] = v                    
                 else: 
-                    obj[k] = f"|{v:0.5g}|" #rounding by any means is considerably slower than this crazy method
+                    #rounding by any means is considerably slower than this crazy method
+                    #we format as a truncated string and then manually remove the string
+                    #indicators from the json via string replace methods
+                    obj[k] = f"|{v:0.5g}|" 
             else:
                 obj[k] = self._min(v)
 
@@ -254,6 +258,8 @@ class ArffReader(Filter[Iterable[str], _T_Data]):
 
         for line in lines:
 
+            line = line.strip()
+
             if in_meta_section:
 
                 if self._r_comment.match(line): continue
@@ -446,7 +452,7 @@ class Structure(Filter[_T_Data, Iterable[Any]]):
 
         key_structure = self._recursive_structure_keys(header, self._col_structure)
 
-        for row in data:
+        for row in data :
            yield self._recursive_structure_rows(row, key_structure) 
 
     def _recursive_structure_keys(self, header, cols):
