@@ -5,7 +5,7 @@ from itertools import chain
 from queue import Queue
 from typing import Iterable, Sequence, TypeVar, Any, Generic, Union
 
-from coba.pipes.core import Sink, Source
+from coba.pipes.primitives import Sink, Source
 
 _T_in  = TypeVar("_T_in", bound=Any, contravariant=True)
 _T_out = TypeVar("_T_out", bound=Any, covariant=True)
@@ -20,14 +20,25 @@ class NullIO(IO[Any,Any]):
     def write(self, item: Any) -> None:
         pass
 
-class ConsoleIO(IO[str,str]):
+class ConsoleIO(IO[str,Union[Any,Iterable[Any]]]):
     def read() -> str:
         return input()
 
-    def write(self, item: str) -> None:
-        print(item)
+    def write(self, item: Union[Any,Iterable[Any]]) -> None:
+        try:
+            if isinstance(item,str):
+                print(item)
+            else:
+               for it in item:
+                   print(it)
+        except TypeError as e:
+            if "not iterable" in str(e):
+                print(item)
+            else:
+                raise
 
 class DiskIO(IO[Iterable[str],Union[str,Iterable[str]]]):
+    
     def __init__(self, filename:str, append=True):
         
         #If you are using the gzip functionality of disk sink
@@ -85,7 +96,8 @@ class MemoryIO(IO[Sequence[Any], Any]):
             if isinstance(item,str):
                 self.items.append(item)
             else:
-                self.items.extend(list(item))
+                for item in item:
+                    self.items.append(item)
         except TypeError as e:
             if "not iterable" in str(e):
                 self.items.append(item)
@@ -133,4 +145,4 @@ class HttpIO(IO[requests.Response, Any]):
         return requests.get(self._url, stream=True) #by default sends accept-encoding gzip and deflate
 
     def write(self, item: Any) -> None:
-        requests.put(self._url, item)
\ No newline at end of file
+        requests.put(self._url, item)
